@@ -169,15 +169,60 @@ static unsigned propnum(const char *name)
 	return offset;
 }
 
+/*
+ * Add ranges by comparing 'base' and 'end' addresses with usable
+ * memory ranges. Returns the number of ranges added. Each range added
+ * increments 'idx' by 2.
+ */
+static uint64_t add_ranges(uint64_t **ranges, int *ranges_size, int idx,
+			   uint64_t base, uint64_t end)
+{
+	uint64_t loc_base, loc_end, rngs_cnt = 0;
+	size_t range;
+	int add = 0;
+
+	for (range = 0; range < usablemem_rgns.size; range++) {
+		loc_base = usablemem_rgns.ranges[range].start;
+		loc_end = usablemem_rgns.ranges[range].end;
+		if (loc_base >= base && loc_end <= end) {
+			add = 1;
+		} else if (base < loc_end && end > loc_base) {
+			if (loc_base < base)
+				loc_base = base;
+			if (loc_end > end)
+				loc_end = end;
+			add = 1;
+		}
+
+		if (add) {
+			if (idx >= ((*ranges_size) - 2)) {
+				(*ranges_size) += MEM_RANGE_CHUNK_SZ;
+				*ranges = realloc(*ranges, (*ranges_size)*8);
+				if (!(*ranges))
+					die("unrecoverable error: can't realloc"
+					    "%d bytes for ranges.\n",
+					    (*ranges_size)*8);
+			}
+			(*ranges)[idx++] = cpu_to_be64(loc_base);
+			(*ranges)[idx++] = cpu_to_be64(loc_end - loc_base);
+
+			rngs_cnt++;
+		}
+	}
+
+	return rngs_cnt;
+}
+
 #ifdef HAVE_DYNAMIC_MEMORY
 static void add_dyn_reconf_usable_mem_property__(int fd)
 {
 	char fname[MAXPATH], *bname;
-	uint64_t buf[32];
+	char buf[32];
+	uint32_t lmbs_in_set = 1;
 	uint64_t *ranges;
 	int ranges_size = MEM_RANGE_CHUNK_SZ;
-	uint64_t base, end, loc_base, loc_end;
-	size_t i, rngs_cnt, range;
+	uint64_t base, end, rngs_cnt;
+	size_t i, j;
 	int rlen = 0;
 	int tmp_indx;
 
@@ -198,35 +243,46 @@ static void add_dyn_reconf_usable_mem_property__(int fd)
 		    ranges_size*8);
 
 	rlen = 0;
-	for (i = 0; i < num_of_lmbs; i++) {
-		if (read(fd, buf, 24) < 0)
+	for (i = 0; i < num_of_lmb_sets; i++) {
+		if (read(fd, buf, LMB_ENTRY_SIZE) < 0)
 			die("unrecoverable error: error reading \"%s\": %s\n",
 				pathname, strerror(errno));
 
-		base = be64_to_cpu((uint64_t) buf[0]);
-		end = base + lmb_size;
-		if (~0ULL - base < end)
-			die("unrecoverable error: mem property overflow\n");
+		/*
+		 * If the property is ibm,dynamic-memory-v2, the first 4 bytes
+		 * tell the number of sequential LMBs in this entry. Else, if
+		 * the property is ibm,dynamic-memory, each entry represents
+		 * one LMB. Make sure to add an entry for each LMB as kernel
+		 * looks for a counter for every LMB.
+		 */
+		if (is_dyn_mem_v2)
+			lmbs_in_set = be32_to_cpu(((unsigned int *)buf)[0]);
 
-		tmp_indx = rlen++;
-
-		rngs_cnt = 0;
-		for (range = 0; range < usablemem_rgns.size; range++) {
-			int add = 0;
-			loc_base = usablemem_rgns.ranges[range].start;
-			loc_end = usablemem_rgns.ranges[range].end;
-			if (loc_base >= base && loc_end <= end) {
-				add = 1;
-			} else if (base < loc_end && end > loc_base) {
-				if (loc_base < base)
-					loc_base = base;
-				if (loc_end > end)
-					loc_end = end;
-				add = 1;
+		base = be64_to_cpu(*((uint64_t *)&buf[DRCONF_ADDR]));
+		for (j = 0; j < lmbs_in_set; j++) {
+			end = base + lmb_size;
+			if (~0ULL - base < end) {
+				die("unrecoverable error: mem property"
+				    " overflow\n");
 			}
 
-			if (add) {
-				if (rlen >= (ranges_size-2)) {
+			tmp_indx = rlen++;
+
+			rngs_cnt = add_ranges(&ranges, &ranges_size, rlen,
+					      base, end);
+			if (rngs_cnt == 0) {
+				/* We still need to add a counter for every LMB
+				 * because the kernel parsing code is dumb. We
+				 * just have a zero in this case, with no
+				 * following base/len.
+				 */
+				ranges[tmp_indx] = 0;
+
+				/* rlen is already just tmp_indx+1 as we didn't
+				 * write anything. Check array size here, as we
+				 * will probably go on writing zeros for a while
+				 */
+				if (rlen >= (ranges_size-1)) {
 					ranges_size += MEM_RANGE_CHUNK_SZ;
 					ranges = realloc(ranges, ranges_size*8);
 					if (!ranges)
@@ -235,33 +291,13 @@ static void add_dyn_reconf_usable_mem_property__(int fd)
 						    " ranges.\n",
 						    ranges_size*8);
 				}
-				ranges[rlen++] = cpu_to_be64(loc_base);
-				ranges[rlen++] = cpu_to_be64(loc_end - loc_base);
-				rngs_cnt++;
+			} else {
+				/* Store the count of (base, size) duple */
+				ranges[tmp_indx] = cpu_to_be64(rngs_cnt);
+				rlen += rngs_cnt * 2;
 			}
-		}
-		if (rngs_cnt == 0) {
-			/* We still need to add a counter for every LMB because
-			 * the kernel parsing code is dumb.  We just have
-			 * a zero in this case, with no following base/len.
-			 */
-			ranges[tmp_indx] = 0;
-			/* rlen is already just tmp_indx+1 as we didn't write
-			 * anything.  Check array size here, as we'll probably
-			 * go on for a while writing zeros now.
-			 */
-			if (rlen >= (ranges_size-1)) {
-				ranges_size += MEM_RANGE_CHUNK_SZ;
-				ranges = realloc(ranges, ranges_size*8);
-				if (!ranges)
-					die("unrecoverable error: can't"
-					    " realloc %d bytes for"
-					    " ranges.\n",
-					    ranges_size*8);
-			}
-		} else {
-			/* Store the count of (base, size) duple */
-			ranges[tmp_indx] = cpu_to_be64((uint64_t) rngs_cnt);
+
+			base = end;
 		}
 	}
 		
@@ -281,7 +317,8 @@ static void add_dyn_reconf_usable_mem_property__(int fd)
 
 static void add_dyn_reconf_usable_mem_property(struct dirent *dp, int fd)
 {
-	if (!strcmp(dp->d_name, "ibm,dynamic-memory") && usablemem_rgns.size)
+	if ((!strcmp(dp->d_name, "ibm,dynamic-memory-v2") ||
+	    !strcmp(dp->d_name, "ibm,dynamic-memory")) && usablemem_rgns.size)
 		add_dyn_reconf_usable_mem_property__(fd);
 }
 #else
@@ -294,8 +331,7 @@ static void add_usable_mem_property(int fd, size_t len)
 	uint64_t buf[2];
 	uint64_t *ranges;
 	int ranges_size = MEM_RANGE_CHUNK_SZ;
-	uint64_t base, end, loc_base, loc_end;
-	size_t range;
+	uint64_t base, end, rngs_cnt;
 	int rlen = 0;
 
 	strcpy(fname, pathname);
@@ -326,33 +362,8 @@ static void add_usable_mem_property(int fd, size_t len)
 		die("unrecoverable error: can't alloc %zu bytes for ranges.\n",
 		    ranges_size * sizeof(*ranges));
 
-	for (range = 0; range < usablemem_rgns.size; range++) {
-		int add = 0;
-		loc_base = usablemem_rgns.ranges[range].start;
-		loc_end = usablemem_rgns.ranges[range].end;
-		if (loc_base >= base && loc_end <= end) {
-			add = 1;
-		} else if (base < loc_end && end > loc_base) {
-			if (loc_base < base)
-				loc_base = base;
-			if (loc_end > end)
-				loc_end = end;
-			add = 1;
-		}
-		if (add) {
-			if (rlen >= (ranges_size-2)) {
-				ranges_size += MEM_RANGE_CHUNK_SZ;
-				ranges = realloc(ranges, ranges_size *
-						 sizeof(*ranges));
-				if (!ranges)
-					die("unrecoverable error: can't realloc"
-					    "%zu bytes for ranges.\n",
-					    ranges_size*sizeof(*ranges));
-			}
-			ranges[rlen++] = cpu_to_be64(loc_base);
-			ranges[rlen++] = cpu_to_be64(loc_end - loc_base);
-		}
-	}
+	rngs_cnt = add_ranges(&ranges, &ranges_size, rlen, base, end);
+	rlen += rngs_cnt * 2;
 
 	if (!rlen) {
 		/*
@@ -524,19 +535,21 @@ static void dt_copy_old_root_param(void)
 	strcpy(filename, pathname);
 	strcat(filename, "bootargs");
 	fp = fopen(filename, "r");
-	if (fp) {
-		if (getline(&last_cmdline, &len, fp) == -1)
-			die("unable to read %s\n", filename);
+	if (!fp)
+		return;
 
-		p = strstr(last_cmdline, "root=");
-		if (p) {
-			old_param = strtok(p, " ");
-			len = strlen(local_cmdline);
-			if (len != 0)
-				strcat(local_cmdline, " ");
-			strcat(local_cmdline, old_param);
-		}
+	if (getline(&last_cmdline, &len, fp) == -1)
+		die("unable to read %s\n", filename);
+
+	p = strstr(last_cmdline, "root=");
+	if (p) {
+		old_param = strtok(p, " ");
+		len = strlen(local_cmdline);
+		if (len != 0)
+			strcat(local_cmdline, " ");
+		strcat(local_cmdline, old_param);
 	}
+
 	if (last_cmdline)
 		free(last_cmdline);
 
